@@ -1,20 +1,28 @@
 use async_trait::async_trait;
+use serde_json::Value;
+use std::collections::BTreeMap;
 use thiserror::Error;
 
-use ssi::did::{DIDMethod, Document, Source, VerificationMethod, VerificationMethodMap, DIDURL};
+use ssi::did::{
+    Context, Contexts, DIDMethod, Document, Source, VerificationMethod, VerificationMethodMap,
+    DEFAULT_CONTEXT, DIDURL,
+};
 use ssi::did_resolve::{
     DIDResolver, DocumentMetadata, ResolutionInputMetadata, ResolutionMetadata, ERROR_INVALID_DID,
-    ERROR_NOT_FOUND, TYPE_DID_LD_JSON,
+    ERROR_NOT_FOUND,
 };
 #[cfg(feature = "secp256r1")]
 use ssi::jwk::p256_parse;
+use ssi::jwk::rsa_x509_pub_parse;
 #[cfg(feature = "secp256k1")]
 use ssi::jwk::secp256k1_parse;
 use ssi::jwk::{Base64urlUInt, OctetParams, Params, JWK};
 
 const DID_KEY_ED25519_PREFIX: [u8; 2] = [0xed, 0x01];
 const DID_KEY_SECP256K1_PREFIX: [u8; 2] = [0xe7, 0x01];
+const DID_KEY_BLS12381_G2_PREFIX: [u8; 2] = [0xeb, 0x01];
 const DID_KEY_P256_PREFIX: [u8; 2] = [0x80, 0x24];
+const DID_KEY_RSA_PREFIX: [u8; 2] = [0x85, 0x24];
 
 #[derive(Error, Debug)]
 pub enum DIDKeyError {
@@ -41,6 +49,7 @@ impl DIDResolver for DIDKey {
         Option<DocumentMetadata>,
     ) {
         let vm_type;
+        let vm_type_iri;
         if !did.starts_with("did:key:") {
             return (
                 ResolutionMetadata {
@@ -79,6 +88,15 @@ impl DIDResolver for DIDKey {
                 None,
             );
         }
+        let mut context = BTreeMap::new();
+        context.insert(
+            "publicKeyJwk".to_string(),
+            serde_json::json!({
+                "@id": "https://w3id.org/security#publicKeyJwk",
+                "@type": "@json"
+            }),
+        );
+
         let jwk = if data[0] == DID_KEY_ED25519_PREFIX[0] && data[1] == DID_KEY_ED25519_PREFIX[1] {
             if data.len() - 2 != 32 {
                 return (
@@ -92,6 +110,7 @@ impl DIDResolver for DIDKey {
                 );
             }
             vm_type = "Ed25519VerificationKey2018".to_string();
+            vm_type_iri = "https://w3id.org/security#Ed25519VerificationKey2018".to_string();
             JWK {
                 params: Params::OKP(OctetParams {
                     curve: "Ed25519".to_string(),
@@ -119,6 +138,8 @@ impl DIDResolver for DIDKey {
             match secp256k1_parse(&data[2..]) {
                 Ok(jwk) => {
                     vm_type = "EcdsaSecp256k1VerificationKey2019".to_string();
+                    vm_type_iri =
+                        "https://w3id.org/security#EcdsaSecp256k1VerificationKey2019".to_string();
                     jwk
                 }
                 Err(err) => return (ResolutionMetadata::from_error(&err), None, None),
@@ -134,6 +155,8 @@ impl DIDResolver for DIDKey {
             match p256_parse(&data[2..]) {
                 Ok(jwk) => {
                     vm_type = "EcdsaSecp256r1VerificationKey2019".to_string();
+                    vm_type_iri =
+                        "https://w3id.org/security#EcdsaSecp256r1VerificationKey2019".to_string();
                     jwk
                 }
                 Err(err) => return (ResolutionMetadata::from_error(&err.to_string()), None, None),
@@ -144,6 +167,35 @@ impl DIDResolver for DIDKey {
                 None,
                 None,
             );
+        } else if data[0] == DID_KEY_RSA_PREFIX[0] && data[1] == DID_KEY_RSA_PREFIX[1] {
+            match rsa_x509_pub_parse(&data[2..]) {
+                Ok(jwk) => {
+                    vm_type = "JsonWebKey2020".to_string();
+                    vm_type_iri = "https://w3id.org/security#JsonWebKey2020".to_string();
+                    jwk
+                }
+                Err(err) => return (ResolutionMetadata::from_error(&err.to_string()), None, None),
+            }
+        } else if data[0] == DID_KEY_BLS12381_G2_PREFIX[0]
+            && data[1] == DID_KEY_BLS12381_G2_PREFIX[1]
+        {
+            {
+                if data.len() - 2 != 96 {
+                    return (
+                        ResolutionMetadata::from_error(ERROR_INVALID_DID),
+                        None,
+                        None,
+                    );
+                }
+                vm_type = "Bls12381G2Key2020".to_string();
+                vm_type_iri = "https://w3id.org/security#Bls12381G2Key2020".to_string();
+                // https://datatracker.ietf.org/doc/html/draft-denhartog-pairing-curves-jose-cose-00#section-3.1.3
+                JWK::from(Params::OKP(OctetParams {
+                    curve: "Bls12381G2".to_string(),
+                    public_key: Base64urlUInt(data[2..].to_vec()),
+                    private_key: None,
+                }))
+            }
         } else {
             return (
                 ResolutionMetadata {
@@ -155,30 +207,31 @@ impl DIDResolver for DIDKey {
                 None,
             );
         };
+        context.insert(vm_type.to_string(), Value::String(vm_type_iri));
         let vm_didurl = DIDURL {
             did: did.to_string(),
             fragment: Some(method_specific_id.to_string()),
             ..Default::default()
         };
         let doc = Document {
+            context: Contexts::Many(vec![
+                Context::URI(DEFAULT_CONTEXT.to_string()),
+                Context::Object(context),
+            ]),
             id: did.to_string(),
             verification_method: Some(vec![VerificationMethod::Map(VerificationMethodMap {
-                id: did.to_string() + &"#" + method_specific_id,
+                id: format!("{}#{}", did, method_specific_id),
                 type_: vm_type,
                 controller: did.to_string(),
                 public_key_jwk: Some(jwk),
                 ..Default::default()
             })]),
             authentication: Some(vec![VerificationMethod::DIDURL(vm_didurl.clone())]),
-            assertion_method: Some(vec![VerificationMethod::DIDURL(vm_didurl.clone())]),
+            assertion_method: Some(vec![VerificationMethod::DIDURL(vm_didurl)]),
             ..Default::default()
         };
         (
-            ResolutionMetadata {
-                error: None,
-                content_type: Some(TYPE_DID_LD_JSON.to_string()),
-                property_set: None,
-            },
+            ResolutionMetadata::default(),
             Some(doc),
             Some(DocumentMetadata::default()),
         )
@@ -187,7 +240,7 @@ impl DIDResolver for DIDKey {
 
 impl DIDMethod for DIDKey {
     fn name(&self) -> &'static str {
-        return "key";
+        "key"
     }
 
     fn generate(&self, source: &Source) -> Option<String> {
@@ -213,6 +266,17 @@ impl DIDMethod for DIDKey {
                                     .concat(),
                             )
                     }
+                    "Bls12381G2" => {
+                        "did:key:".to_string()
+                            + &multibase::encode(
+                                multibase::Base::Base58Btc,
+                                [
+                                    DID_KEY_BLS12381_G2_PREFIX.to_vec(),
+                                    params.public_key.0.clone(),
+                                ]
+                                .concat(),
+                            )
+                    }
                     //_ => return Some(Err(DIDKeyError::UnsupportedCurve(params.curve.clone()))),
                     _ => return None,
                 }
@@ -243,22 +307,34 @@ impl DIDMethod for DIDKey {
                     }
                     #[cfg(feature = "secp256r1")]
                     "P-256" => {
-                        // P-256 did:key is uncompressed
-                        let (x, y) =
-                            match (params.x_coordinate.as_ref(), params.y_coordinate.as_ref()) {
-                                (Some(x), Some(y)) => (x.0.to_vec(), y.0.to_vec()),
-                                _ => return None,
-                            };
+                        use p256::elliptic_curve::sec1::ToEncodedPoint;
+                        use std::convert::TryFrom;
+                        let pk = match p256::PublicKey::try_from(params) {
+                            Ok(pk) => pk,
+                            Err(_err) => return None,
+                        };
                         "did:key:".to_string()
                             + &multibase::encode(
                                 multibase::Base::Base58Btc,
-                                [DID_KEY_P256_PREFIX.to_vec(), x, y].concat(),
+                                [
+                                    DID_KEY_P256_PREFIX.to_vec(),
+                                    pk.to_encoded_point(true).as_bytes().to_vec(),
+                                ]
+                                .concat(),
                             )
                     }
                     //_ => return Some(Err(DIDKeyError::UnsupportedCurve(params.curve.clone()))),
                     _ => return None,
                 }
             }
+            Params::RSA(ref params) => {
+                let der = simple_asn1::der_encode(&params.to_public()).ok()?;
+                "did:key:".to_string()
+                    + &multibase::encode(
+                        multibase::Base::Base58Btc,
+                        [DID_KEY_RSA_PREFIX.to_vec(), der.to_vec()].concat(),
+                    )
+            }
             _ => return None, // _ => return Some(Err(DIDKeyError::UnsupportedKeyType)),
         };
         Some(did)
@@ -316,13 +392,13 @@ mod tests {
     #[async_std::test]
     async fn from_did_key_p256() {
         // https://w3c-ccg.github.io/did-method-key/#p-256
-        let did = "did:key:zrurwcJZss4ruepVNu1H3xmSirvNbzgBk9qrCktB6kaewXnJAhYWwtP3bxACqBpzjZdN7TyHNzzGGSSH5qvZsSDir9z";
+        let did = "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169";
         let (res_meta, _doc, _doc_meta) = DIDKey
             .resolve(did, &ResolutionInputMetadata::default())
             .await;
         assert_eq!(res_meta.error, None);
 
-        let vm = "did:key:zrurwcJZss4ruepVNu1H3xmSirvNbzgBk9qrCktB6kaewXnJAhYWwtP3bxACqBpzjZdN7TyHNzzGGSSH5qvZsSDir9z#zrurwcJZss4ruepVNu1H3xmSirvNbzgBk9qrCktB6kaewXnJAhYWwtP3bxACqBpzjZdN7TyHNzzGGSSH5qvZsSDir9z";
+        let vm = "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169#zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169";
         let (res_meta, object, _meta) =
             dereference(&DIDKey, &vm, &DereferencingInputMetadata::default()).await;
         assert_eq!(res_meta.error, None);
@@ -347,6 +423,72 @@ mod tests {
         assert_eq!(did1, did);
     }
 
+    #[async_std::test]
+    async fn from_did_key_bls() {
+        // https://w3c-ccg.github.io/did-method-key/#bls-12381
+        let did = "did:key:zUC7K4ndUaGZgV7Cp2yJy6JtMoUHY6u7tkcSYUvPrEidqBmLCTLmi6d5WvwnUqejscAkERJ3bfjEiSYtdPkRSE8kSa11hFBr4sTgnbZ95SJj19PN2jdvJjyzpSZgxkyyxNnBNnY";
+        let (res_meta, _doc, _doc_meta) = DIDKey
+            .resolve(did, &ResolutionInputMetadata::default())
+            .await;
+        assert_eq!(res_meta.error, None);
+
+        let vm = "did:key:zUC7K4ndUaGZgV7Cp2yJy6JtMoUHY6u7tkcSYUvPrEidqBmLCTLmi6d5WvwnUqejscAkERJ3bfjEiSYtdPkRSE8kSa11hFBr4sTgnbZ95SJj19PN2jdvJjyzpSZgxkyyxNnBNnY#zUC7K4ndUaGZgV7Cp2yJy6JtMoUHY6u7tkcSYUvPrEidqBmLCTLmi6d5WvwnUqejscAkERJ3bfjEiSYtdPkRSE8kSa11hFBr4sTgnbZ95SJj19PN2jdvJjyzpSZgxkyyxNnBNnY";
+        let (res_meta, object, _meta) =
+            dereference(&DIDKey, &vm, &DereferencingInputMetadata::default()).await;
+        assert_eq!(res_meta.error, None);
+        let vm = match object {
+            Content::Object(Resource::VerificationMethod(vm)) => vm,
+            _ => unreachable!(),
+        };
+        let key = vm.public_key_jwk.unwrap();
+        eprintln!("key {}", serde_json::to_string_pretty(&key).unwrap());
+
+        // Note: this "x" value is generated by this implementation - not yet confirmed with other
+        // implementations.
+        // Related issue: https://github.com/mattrglobal/bls12381-jwk-draft/issues/5
+        let key_expected: JWK = serde_json::from_value(serde_json::json!({
+            "kty": "OKP",
+            "crv": "Bls12381G2",
+            "x": "tKWJu0SOY7onl4tEyOOH11XBriQN2JgzV-UmjgBMSsNkcAx3_l97SVYViSDBouTVBkBfrLh33C5icDD-4UEDxNO3Wn1ijMHvn2N63DU4pkezA3kGN81jGbwbrsMPpiOF"
+        }))
+        .unwrap();
+        assert_eq!(key, key_expected);
+
+        let did1 = DIDKey.generate(&Source::Key(&key)).unwrap();
+        assert_eq!(did1, did);
+    }
+
+    #[async_std::test]
+    async fn from_did_key_rsa() {
+        let did = "did:key:z4MXj1wBzi9jUstyPMS4jQqB6KdJaiatPkAtVtGc6bQEQEEsKTic4G7Rou3iBf9vPmT5dbkm9qsZsuVNjq8HCuW1w24nhBFGkRE4cd2Uf2tfrB3N7h4mnyPp1BF3ZttHTYv3DLUPi1zMdkULiow3M1GfXkoC6DoxDUm1jmN6GBj22SjVsr6dxezRVQc7aj9TxE7JLbMH1wh5X3kA58H3DFW8rnYMakFGbca5CB2Jf6CnGQZmL7o5uJAdTwXfy2iiiyPxXEGerMhHwhjTA1mKYobyk2CpeEcmvynADfNZ5MBvcCS7m3XkFCMNUYBS9NQ3fze6vMSUPsNa6GVYmKx2x6JrdEjCk3qRMMmyjnjCMfR4pXbRMZa3i";
+        let (res_meta, _doc, _doc_meta) = DIDKey
+            .resolve(did, &ResolutionInputMetadata::default())
+            .await;
+        assert_eq!(res_meta.error, None);
+
+        let vm = "did:key:z4MXj1wBzi9jUstyPMS4jQqB6KdJaiatPkAtVtGc6bQEQEEsKTic4G7Rou3iBf9vPmT5dbkm9qsZsuVNjq8HCuW1w24nhBFGkRE4cd2Uf2tfrB3N7h4mnyPp1BF3ZttHTYv3DLUPi1zMdkULiow3M1GfXkoC6DoxDUm1jmN6GBj22SjVsr6dxezRVQc7aj9TxE7JLbMH1wh5X3kA58H3DFW8rnYMakFGbca5CB2Jf6CnGQZmL7o5uJAdTwXfy2iiiyPxXEGerMhHwhjTA1mKYobyk2CpeEcmvynADfNZ5MBvcCS7m3XkFCMNUYBS9NQ3fze6vMSUPsNa6GVYmKx2x6JrdEjCk3qRMMmyjnjCMfR4pXbRMZa3i#z4MXj1wBzi9jUstyPMS4jQqB6KdJaiatPkAtVtGc6bQEQEEsKTic4G7Rou3iBf9vPmT5dbkm9qsZsuVNjq8HCuW1w24nhBFGkRE4cd2Uf2tfrB3N7h4mnyPp1BF3ZttHTYv3DLUPi1zMdkULiow3M1GfXkoC6DoxDUm1jmN6GBj22SjVsr6dxezRVQc7aj9TxE7JLbMH1wh5X3kA58H3DFW8rnYMakFGbca5CB2Jf6CnGQZmL7o5uJAdTwXfy2iiiyPxXEGerMhHwhjTA1mKYobyk2CpeEcmvynADfNZ5MBvcCS7m3XkFCMNUYBS9NQ3fze6vMSUPsNa6GVYmKx2x6JrdEjCk3qRMMmyjnjCMfR4pXbRMZa3i";
+        let (res_meta, object, _meta) =
+            dereference(&DIDKey, &vm, &DereferencingInputMetadata::default()).await;
+        assert_eq!(res_meta.error, None);
+        let vm = match object {
+            Content::Object(Resource::VerificationMethod(vm)) => vm,
+            _ => unreachable!(),
+        };
+        let key = vm.public_key_jwk.unwrap();
+        eprintln!("key {}", serde_json::to_string_pretty(&key).unwrap());
+
+        let key_expected: JWK = serde_json::from_value(serde_json::json!({
+            "kty": "RSA",
+            "e": "AQAB",
+            "n": "sbX82NTV6IylxCh7MfV4hlyvaniCajuP97GyOqSvTmoEdBOflFvZ06kR_9D6ctt45Fk6hskfnag2GG69NALVH2o4RCR6tQiLRpKcMRtDYE_thEmfBvDzm_VVkOIYfxu-Ipuo9J_S5XDNDjczx2v-3oDh5-CIHkU46hvFeCvpUS-L8TJSbgX0kjVk_m4eIb9wh63rtmD6Uz_KBtCo5mmR4TEtcLZKYdqMp3wCjN-TlgHiz_4oVXWbHUefCEe8rFnX1iQnpDHU49_SaXQoud1jCaexFn25n-Aa8f8bc5Vm-5SeRwidHa6ErvEhTvf1dz6GoNPp2iRvm-wJ1gxwWJEYPQ"
+        }))
+        .unwrap();
+        assert_eq!(key, key_expected);
+
+        let did1 = DIDKey.generate(&Source::Key(&key)).unwrap();
+        assert_eq!(did1, did);
+    }
+
     #[async_std::test]
     async fn credential_prove_verify_did_key() {
         use ssi::vc::{get_verification_method, Credential, Issuer, LinkedDataProofOptions, URI};
@@ -367,8 +509,11 @@ mod tests {
         let verification_method = get_verification_method(&did, &DIDKey).await.unwrap();
         let mut issue_options = LinkedDataProofOptions::default();
         vc.issuer = Some(Issuer::URI(URI::String(did.clone())));
-        issue_options.verification_method = Some(verification_method);
-        let proof = vc.generate_proof(&key, &issue_options).await.unwrap();
+        issue_options.verification_method = Some(URI::String(verification_method));
+        let proof = vc
+            .generate_proof(&key, &issue_options, &DIDKey)
+            .await
+            .unwrap();
         println!("{}", serde_json::to_string_pretty(&proof).unwrap());
         vc.add_proof(proof);
         vc.validate().unwrap();
@@ -402,8 +547,11 @@ mod tests {
 
         let verification_method = get_verification_method(&did, &DIDKey).await.unwrap();
         let mut issue_options = LinkedDataProofOptions::default();
-        issue_options.verification_method = Some(verification_method);
-        let proof = vc.generate_proof(&key, &issue_options).await.unwrap();
+        issue_options.verification_method = Some(URI::String(verification_method));
+        let proof = vc
+            .generate_proof(&key, &issue_options, &DIDKey)
+            .await
+            .unwrap();
         println!("{}", serde_json::to_string_pretty(&proof).unwrap());
         vc.add_proof(proof);
         vc.validate().unwrap();
@@ -437,8 +585,11 @@ mod tests {
 
         let verification_method = get_verification_method(&did, &DIDKey).await.unwrap();
         let mut issue_options = LinkedDataProofOptions::default();
-        issue_options.verification_method = Some(verification_method);
-        let proof = vc.generate_proof(&key, &issue_options).await.unwrap();
+        issue_options.verification_method = Some(URI::String(verification_method));
+        let proof = vc
+            .generate_proof(&key, &issue_options, &DIDKey)
+            .await
+            .unwrap();
         println!("{}", serde_json::to_string_pretty(&proof).unwrap());
         vc.add_proof(proof);
         vc.validate().unwrap();
